@@ -54,8 +54,6 @@ static int  CC_close_eof_cursors(ConnectionClass *self);
 
 static void LIBPQ_update_transaction_status(ConnectionClass *self);
 
-extern GLOBAL_VALUES globals;
-
 
 static void CC_set_error_if_not_set(ConnectionClass *self, int errornumber, const char *errormsg, const char *func)
 {
@@ -133,12 +131,13 @@ PGAPI_Connect(HDBC hdbc,
 	}
 
 	ci = &conn->connInfo;
-	CC_conninfo_init(ci, COPY_GLOBALS);
+	CC_conninfo_init(ci, INIT_GLOBALS);
 
 	make_string(szDSN, cbDSN, ci->dsn, sizeof(ci->dsn));
 
 	/* get the values for the DSN from the registry */
-	getDSNinfo(ci, CONN_OVERWRITE);
+	getDSNinfo(ci, NULL);
+
 	logs_on_off(1, ci->drivers.debug, ci->drivers.commlog);
 	/* initialize pg_version from connInfo.protocol    */
 	CC_initialize_pg_version(conn);
@@ -159,9 +158,6 @@ PGAPI_Connect(HDBC hdbc,
 		free(tmpstr);
 	}
 
-	/* fill in any defaults */
-	getDSNdefaults(ci);
-
 	qlog("conn = %p, %s(DSN='%s', UID='%s', PWD='%s')\n", conn, func, ci->dsn, ci->username, NAME_IS_VALID(ci->password) ? "xxxxx" : "");
 
 	if ((fchar = CC_connect(conn, NULL)) <= 0)
@@ -452,12 +448,12 @@ CC_examine_global_transaction(ConnectionClass *self)
 }
 
 
-static const char *bgncmd = "BEGIN";
-static const char *cmtcmd = "COMMIT";
-static const char *rbkcmd = "ROLLBACK";
-static const char *svpcmd = "SAVEPOINT";
-static const char *per_query_svp = "_per_query_svp_";
-static const char *rlscmd = "RELEASE";
+CSTR	bgncmd = "BEGIN";
+CSTR	cmtcmd = "COMMIT";
+CSTR	rbkcmd = "ROLLBACK";
+CSTR	svpcmd = "SAVEPOINT";
+CSTR	per_query_svp = "_per_query_svp_";
+CSTR	rlscmd = "RELEASE";
 
 /*
  *	Used to begin a transaction.
@@ -634,8 +630,8 @@ CC_send_client_encoding(ConnectionClass *self, const char * encoding)
 		int	errnum = CC_get_errornumber(self);
 		BOOL	cmd_success;
 
-		snprintf(query, sizeof(query), "set client_encoding to '%s'", encoding);
-		res = CC_send_query(self, query, NULL, IGNORE_ABORT_ON_CONN | ROLLBACK_ON_ERROR, NULL);
+		SPRINTF_FIXED(query, "set client_encoding to '%s'", encoding);
+		res = CC_send_query(self, query, NULL, 0, NULL);
 		cmd_success = QR_command_maybe_successful(res);
 		QR_Destructor(res);
 		CC_set_errornumber(self, errnum);
@@ -817,7 +813,7 @@ handle_pgres_error(ConnectionClass *self, const PGresult *pgres,
 	{
 		char *sqlstate = PQresultErrorField(pgres, PG_DIAG_SQLSTATE);
 		if (sqlstate)
-			strncpy_null(res->sqlstate, sqlstate, sizeof(res->sqlstate));
+			STRCPY_FIXED(res->sqlstate, sqlstate);
 	}
 
 	if (NULL == pgres &&
@@ -879,7 +875,8 @@ handle_pgres_error(ConnectionClass *self, const PGresult *pgres,
 		goto cleanup;
 	}
 
-	mylog("error message len=%d\n", strlen(errmsg));
+	if (get_mylog() > 0)
+		mylog("error message=%s(%d)\n", errmsg, strlen(errmsg));
 
 	if (res)
 	{
@@ -985,12 +982,10 @@ static char CC_initial_log(ConnectionClass *self, const char *func)
 	if (NULL == self->locale_encoding)
 	{
 		encoding = check_client_encoding(ci->conn_settings);
-		if (!encoding)
-			encoding = check_client_encoding(ci->drivers.conn_settings);
 		CC_set_locale_encoding(self, encoding);
 		qlog("                extra_systable_prefixes='%s', conn_settings='%s' conn_encoding='%s'\n",
 			ci->drivers.extra_systable_prefixes,
-			PRINT_NAME(ci->drivers.conn_settings),
+			PRINT_NAME(ci->conn_settings),
 			encoding ? encoding : "");
 	}
 	if (self->status == CONN_DOWN)
@@ -1028,7 +1023,7 @@ LIBPQ_CC_connect(ConnectionClass *self, char *salt_para)
 
 	if (ret = LIBPQ_connect(self), ret <= 0)
 		return ret;
-	res = CC_send_query(self, "SET DateStyle = 'ISO';SET extra_float_digits = 2;" ISOLATION_SHOW_QUERY, NULL, 0, NULL);
+	res = CC_send_query(self, "SET DateStyle = 'ISO';SET extra_float_digits = 2;" ISOLATION_SHOW_QUERY, NULL, READ_ONLY_QUERY, NULL);
 	if (QR_command_maybe_successful(res))
 	{
 		handle_show_results(res);
@@ -1069,11 +1064,8 @@ CC_connect(ConnectionClass *self, char *salt_para)
 	 * function instead.
 	 */
 
-	/* Global settings */
-	retsend = CC_send_settings(self, GET_NAME(self->connInfo.drivers.conn_settings));
 	/* Per Datasource settings */
-	if (retsend)
-		retsend = CC_send_settings(self, GET_NAME(self->connInfo.conn_settings));
+	retsend = CC_send_settings(self, GET_NAME(self->connInfo.conn_settings));
 
 	if (CC_get_errornumber(self) > 0)
 		saverr = strdup(CC_get_errormsg(self));
@@ -1115,14 +1107,7 @@ CC_connect(ConnectionClass *self, char *salt_para)
 			goto cleanup;
 		}
 
-	ci->updatable_cursors = DISALLOW_UPDATABLE_CURSORS;
-	if (ci->allow_keyset)
-	{
-		if (ci->drivers.lie || !ci->drivers.use_declarefetch)
-			ci->updatable_cursors |= (ALLOW_STATIC_CURSORS | ALLOW_KEYSET_DRIVEN_CURSORS | ALLOW_BULK_OPERATIONS | SENSE_SELF_OPERATIONS);
-		else
-			ci->updatable_cursors |= (ALLOW_STATIC_CURSORS | ALLOW_BULK_OPERATIONS | SENSE_SELF_OPERATIONS);
-	}
+	ci_updatable_cursors_set(ci);
 
 	if (CC_get_errornumber(self) > 0)
 		CC_clear_error(self);		/* clear any initial command errors */
@@ -1258,7 +1243,7 @@ int	CC_get_max_idlen(ConnectionClass *self)
 	{
 		QResultClass	*res;
 
-		res = CC_send_query(self, "show max_identifier_length", NULL, ROLLBACK_ON_ERROR | IGNORE_ABORT_ON_CONN, NULL);
+		res = CC_send_query(self, "show max_identifier_length", NULL, READ_ONLY_QUERY, NULL);
 		if (QR_command_maybe_successful(res))
 			len = self->max_identifier_length = QR_get_value_backend_int(res, 0, 0, FALSE);
 		QR_Destructor(res);
@@ -1318,7 +1303,7 @@ SQLUINTEGER	CC_get_isolation(ConnectionClass *self)
 	SQLUINTEGER	isolation = 0;
 	QResultClass	*res;
 
-	res = CC_send_query(self, ISOLATION_SHOW_QUERY, NULL, ROLLBACK_ON_ERROR | IGNORE_ABORT_ON_CONN, NULL);
+	res = CC_send_query(self, ISOLATION_SHOW_QUERY, NULL, READ_ONLY_QUERY, NULL);
 	if (QR_command_maybe_successful(res))
 	{
 		handle_show_results(res);
@@ -1444,16 +1429,18 @@ static void CC_clear_cursors(ConnectionClass *self, BOOL on_abort)
 
 				if (QR_needs_survival_check(res))
 				{
-					snprintf(cmd, sizeof(cmd), "MOVE 0 in \"%s\"", QR_get_cursor(res));
+					SPRINTF_FIXED(cmd, "MOVE 0 in \"%s\"", QR_get_cursor(res));
 					CONNLOCK_RELEASE(self);
-					wres = CC_send_query(self, cmd, NULL, ROLLBACK_ON_ERROR | IGNORE_ABORT_ON_CONN, NULL);
+					wres = CC_send_query(self, cmd, NULL, ROLLBACK_ON_ERROR | IGNORE_ABORT_ON_CONN | READ_ONLY_QUERY, NULL);
 					QR_set_no_survival_check(res);
-					if (QR_command_maybe_successful(wres))
+					if (QR_command_maybe_successful(wres) &&
+					    CONN_ERROR_IGNORED != CC_get_errornumber(self))
 						QR_set_permanent(res);
 					else
 						QR_set_cursor(res, NULL);
 					QR_Destructor(wres);
 					CONNLOCK_ACQUIRE(self);
+inolog(" !!!! %s:%p->permanent -> %d %p\n", __FUNCTION__, res, QR_is_permanent(res), QR_get_cursor(res));
 				}
 				else
 					QR_set_permanent(res);
@@ -1486,12 +1473,17 @@ static void CC_mark_cursors_doubtful(ConnectionClass *self)
 
 void	CC_on_commit(ConnectionClass *conn)
 {
+	if (conn->on_commit_in_progress)
+		return;
+	conn->on_commit_in_progress = 1;
 	CONNLOCK_ACQUIRE(conn);
 	if (CC_is_in_trans(conn))
 	{
 		CC_set_no_trans(conn);
 		CC_set_no_manual_trans(conn);
 	}
+	CC_svp_init(conn);
+	CC_start_stmt(conn);
 	CC_clear_cursors(conn, FALSE);
 	CONNLOCK_RELEASE(conn);
 	CC_discard_marked_objects(conn);
@@ -1504,6 +1496,7 @@ void	CC_on_commit(ConnectionClass *conn)
 		conn->result_uncommitted = 0;
 	}
 	CONNLOCK_RELEASE(conn);
+	conn->on_commit_in_progress = 0;
 }
 void	CC_on_abort(ConnectionClass *conn, UDWORD opt)
 {
@@ -1522,6 +1515,8 @@ mylog("CC_on_abort in opt=%x\n", opt);
 			set_no_trans = TRUE;
 		}
 	}
+	CC_svp_init(conn);
+	CC_start_stmt(conn);
 	CC_clear_cursors(conn, TRUE);
 	if (0 != (opt & CONN_DEAD))
 	{
@@ -1553,8 +1548,8 @@ mylog("CC_on_abort in opt=%x\n", opt);
 void	CC_on_abort_partial(ConnectionClass *conn)
 {
 mylog("CC_on_abort_partial in\n");
-	ProcessRollback(conn, TRUE, TRUE);
 	CONNLOCK_ACQUIRE(conn);
+	ProcessRollback(conn, TRUE, TRUE);
 	CC_discard_marked_objects(conn);
 	CONNLOCK_RELEASE(conn);
 }
@@ -1605,6 +1600,67 @@ CC_from_PGresult(QResultClass *res, StatementClass *stmt,
 	return success;
 }
 
+int
+CC_internal_rollback(ConnectionClass *self, int rollback_type, BOOL ignore_abort)
+{
+	int	ret = 0;
+	char		cmd[64];
+	PGresult   *pgres = NULL;
+
+	if (!CC_is_in_error_trans(self))
+		return 1;
+	switch (rollback_type)
+	{
+		case PER_STATEMENT_ROLLBACK:
+			GenerateSvpCommand(self, INTERNAL_ROLLBACK_OPERATION, cmd, sizeof(cmd));
+			mylog(" %s:rollback_type=%d %s\n", __FUNCTION__, rollback_type, cmd);
+			pgres = PQexec(self->pqconn, cmd);
+			switch (PQresultStatus(pgres))
+			{
+				case PGRES_COMMAND_OK:
+				case PGRES_NONFATAL_ERROR:
+					ret = 1;
+					if (ignore_abort)
+						CC_set_no_error_trans(self);
+					LIBPQ_update_transaction_status(self);
+					break;
+				default:
+					break;
+			}
+			break;
+		case PER_QUERY_ROLLBACK:
+			SPRINTF_FIXED(cmd, "%s TO %s;%s %s"
+				, rbkcmd, per_query_svp , rlscmd, per_query_svp);
+			mylog(" %s:query_rollback PQsendQuery %s\n", __FUNCTION__, cmd);
+			PQsendQuery(self->pqconn, cmd);
+			ret = 1;
+			while (self->pqconn && (pgres = PQgetResult(self->pqconn)) != NULL)
+			{
+				switch (PQresultStatus(pgres))
+				{
+					case PGRES_COMMAND_OK:
+					case PGRES_NONFATAL_ERROR:
+						break;
+					default:
+						ret = 0;
+				}
+			}
+			if (ret)
+			{
+				if (ignore_abort)
+					CC_set_no_error_trans(self);
+			}
+			else
+				mylog(" %s:return error\n", __FUNCTION__);
+			LIBPQ_update_transaction_status(self);
+			break;
+	}
+	if (pgres)
+		PQclear(pgres);
+
+	return ret;
+}
+
 /*
  *	The "result_in" is only used by QR_next_tuple() to fetch another group of rows into
  *	the same existing QResultClass (this occurs when the tuple cache is depleted and
@@ -1630,7 +1686,9 @@ CC_send_query_append(ConnectionClass *self, const char *query, QueryInfo *qi, UD
 	BOOL	ignore_abort_on_conn = ((flag & IGNORE_ABORT_ON_CONN) != 0),
 		create_keyset = ((flag & CREATE_KEYSET) != 0),
 		issue_begin = ((flag & GO_INTO_TRANSACTION) != 0 && !CC_is_in_trans(self)),
-		rollback_on_error, query_rollback, end_with_commit;
+		rollback_on_error, query_rollback, end_with_commit,
+		read_only, prepend_savepoint = FALSE,
+		ignore_roundtrip_time = ((self->connInfo.extra_opts & BIT_IGNORE_ROUND_TRIP_TIME) != 0);
 
 	char		*ptr;
 	BOOL		ReadyToReturn = FALSE,
@@ -1639,11 +1697,11 @@ CC_send_query_append(ConnectionClass *self, const char *query, QueryInfo *qi, UD
 				used_passed_result_object = FALSE,
 			discard_next_begin = FALSE,
 			discard_next_savepoint = FALSE,
+			discard_next_release = FALSE,
 			consider_rollback;
-	int		func_cs_count = 0;
+	int		func_cs_count = 0, i;
 	size_t		query_buf_len = 0;
-	char	   *query_buf = NULL;
-	char	   *query_buf_next;
+	char	   *query_buf = NULL, prepend_cmd[64];
 	size_t		query_len;
 
 	/* QR_set_command() dups this string so doesn't need static */
@@ -1682,7 +1740,7 @@ CC_send_query_append(ConnectionClass *self, const char *query, QueryInfo *qi, UD
 	 *	and the appeneded query would be issued separately.
 	 *	Otherwise a multiple command query would be issued.
 	 */
-	if (appendq && (self->connInfo.extra_opts & BIT_IGNORE_ROUND_TRIP_TIME) != 0)
+	if (appendq && ignore_roundtrip_time)
 	{
 		res = CC_send_query_append(self, query, qi, flag, stmt, NULL);
 		if (QR_command_maybe_successful(res))
@@ -1702,6 +1760,7 @@ CC_send_query_append(ConnectionClass *self, const char *query, QueryInfo *qi, UD
 
 	rollback_on_error = (flag & ROLLBACK_ON_ERROR) != 0;
 	end_with_commit = (flag & END_WITH_COMMIT) != 0;
+	read_only = (flag & READ_ONLY_QUERY) != 0;
 #define	return DONT_CALL_RETURN_FROM_HERE???
 	consider_rollback = (issue_begin || (CC_is_in_trans(self) && !CC_is_in_error_trans(self)) || strnicmp(query, "begin", 5) == 0);
 	if (rollback_on_error)
@@ -1712,9 +1771,15 @@ CC_send_query_append(ConnectionClass *self, const char *query, QueryInfo *qi, UD
 		if (stmt)
 		{
 			StatementClass	*astmt = SC_get_ancestor(stmt);
-			if (!SC_accessed_db(astmt))
+			unsigned int svpopt = 0;
+
+			if (read_only)
+				svpopt |= SVPOPT_RDONLY;
+			if (!ignore_roundtrip_time)
+				svpopt |= SVPOPT_REDUCE_ROUNDTRIP;
+			if (!CC_started_rbpoint(self))
 			{
-				if (SQL_ERROR == SetStatementSvp(astmt))
+				if (SQL_ERROR == SetStatementSvp(astmt, svpopt))
 				{
 					SC_set_error(stmt, STMT_INTERNAL_ERROR, "internal savepoint error", func);
 					goto cleanup;
@@ -1723,69 +1788,78 @@ CC_send_query_append(ConnectionClass *self, const char *query, QueryInfo *qi, UD
 		}
 	}
 
+	/* prepend internal savepoint command ? */
+	if (PREPEND_IN_PROGRESS == self->internal_op)
+		prepend_savepoint = TRUE;
+
 	/* append all these together, to avoid round-trips */
 	query_len = strlen(query);
+	mylog("%s:query_len=%u\n", __FUNCTION__, query_len);
 
-	query_buf_len = strlen(bgncmd) + 1
-		+ strlen(svpcmd) + 1 + strlen(per_query_svp) + 1
-		+ query_len
-		+ (appendq ? (1 + strlen(appendq)) : 0)
-		+ 1 + strlen(rlscmd) + strlen(per_query_svp)
-		+ 1;
-	query_buf = malloc(query_buf_len);
-	if (!query_buf)
+	for (i = 0; i < 2; i++) /* 0:calculate& alloc 1:snprint */
 	{
-		CC_set_error(self, CONN_NO_MEMORY_ERROR, "Couldn't alloc buffer for query.", "");
-		goto cleanup;
-	}
-	query_buf_next = query_buf;
-	if (issue_begin)
-	{
-		strcpy(query_buf_next, bgncmd);
-		query_buf_next += strlen(bgncmd);
-		*(query_buf_next++) = ';';
-		discard_next_begin = TRUE;
-	}
-	if (query_rollback)
-	{
-		strcpy(query_buf_next, svpcmd);
-		query_buf_next += strlen(svpcmd);
-		*(query_buf_next++) = ' ';
-		strcpy(query_buf_next, per_query_svp);
-		query_buf_next += strlen(per_query_svp);
-		*(query_buf_next++) = ';';
-		discard_next_savepoint = TRUE;
-	}
-	memcpy(query_buf_next, query, query_len);
-	query_buf_next += query_len;
-	*query_buf_next = '\0';
-	if (appendq)
-	{
-		*(query_buf_next++) = ';';
-		strcpy(query_buf_next, appendq);
-		query_buf_next += strlen(appendq);
-		*query_buf_next = '\0';
-	}
-	if (query_rollback)
-	{
-		*(query_buf_next++) = ';';
-		strcpy(query_buf_next, rlscmd);
-		query_buf_next += strlen(rlscmd);
-		*(query_buf_next++) = ' ';
-		strcpy(query_buf_next, per_query_svp);
-		query_buf_next += strlen(per_query_svp);
-		*query_buf_next = '\0';
-	}
-
-	if (query_buf_next > query_buf + query_buf_len)
-	{
-		/*
-		 * this should not happen, and if it does, we've already overrun
-		 * the buffer and possibly corrupted memory.
-		 */
-		CC_set_error(self, CONNECTION_COULD_NOT_SEND, "query buffer overrun", func);
-		goto cleanup;
+		/* issue_begin, query_rollback and prepend_savepoint are exclusive */
+		if (issue_begin)
+		{
+			if (0 == i)
+				query_buf_len += (strlen(bgncmd) + 1);
+			else
+			{
+				snprintfcat(query_buf, query_buf_len, "%s;", bgncmd);
+				discard_next_begin = TRUE;
+			}
+		}
+		else if (query_rollback)
+		{
+			if (0 == i)
+				query_buf_len += (strlen(svpcmd) + 1 + strlen(per_query_svp) + 1);
+			else
+			{
+				snprintfcat(query_buf, query_buf_len, "%s %s;", svpcmd, per_query_svp);
+				discard_next_savepoint = TRUE;
+			}
+		}
+		else if (prepend_savepoint)
+		{
+			if (0 == i)
+				query_buf_len += (GenerateSvpCommand(self, INTERNAL_SAVEPOINT_OPERATION, prepend_cmd, sizeof(prepend_cmd)) + 1);
+			else
+			{
+				snprintfcat(query_buf, query_buf_len, "%s;", prepend_cmd);
+				self->internal_op = SAVEPOINT_IN_PROGRESS;
+			}
+		}
+		if (0 == i)
+			query_buf_len += query_len;
+		else
+			strlcat(query_buf, query, query_buf_len);
+		if (appendq)
+		{
+			if (0 == i)
+				query_buf_len += (1 + strlen(appendq));
+			else
+				snprintfcat(query_buf, query_buf_len, ";%s", appendq);
+		}
+		if (query_rollback)
+		{
+			if (0 == i)
+				query_buf_len += (1 + strlen(rlscmd) + 1 + strlen(per_query_svp));
+			else
+				snprintfcat(query_buf, query_buf_len, ";%s %s", rlscmd, per_query_svp);
+		}
+		if (0 == i)
+		{
+			query_buf_len++;
+			query_buf = malloc(query_buf_len);
+			if (!query_buf)
+			{
+				CC_set_error(self, CONN_NO_MEMORY_ERROR, "Couldn't alloc buffer for query.", "");
+				goto cleanup;
+			}
+			query_buf[0] = '\0';
+		}
 	}
+inolog("!!!! %s:query_buf=%s(%d)\n", __FUNCTION__, query_buf, query_buf_len);
 
 	/* Set up notice receiver */
 	nrarg.conn = self;
@@ -1833,6 +1907,7 @@ CC_send_query_append(ConnectionClass *self, const char *query, QueryInfo *qi, UD
 				cmdbuffer = PQcmdStatus(pgres);
 				mylog("send_query: ok - 'C' - %s\n", cmdbuffer);
 
+
 				if (query_completed)	/* allow for "show" style notices */
 				{
 					res->next = QR_Constructor();
@@ -1859,19 +1934,50 @@ CC_send_query_append(ConnectionClass *self, const char *query, QueryInfo *qi, UD
 						break; /* discard the result */
 					}
 				}
+				/*
+				 * There are 2 risks to RELEASE an internal savepoint.
+				 * One is to RELEASE the savepoint invalitated
+				 * due to manually issued ROLLBACK or RELEASE.
+				 * Another is to invalitate manual SAVEPOINTs unexpectedly
+				 * by RELEASing the internal savepoint.
+				 */
 				else if (strnicmp(cmdbuffer, svpcmd, strlen(svpcmd)) == 0)
 				{
 					if (discard_next_savepoint)
 					{
-inolog("Discarded the first SAVEPOINT\n");
 						discard_next_savepoint = FALSE;
+						discard_next_release = TRUE;
+inolog("Discarded a SAVEPOINT result\n");
 						break; /* discard the result */
 					}
+					if (SAVEPOINT_IN_PROGRESS == self->internal_op)
+					{
+						CC_start_rbpoint(self);
+						self->internal_op = 0;
+						break; /* discard the result */
+					}
+					/* Don't release the internal savepoint */
+					self->internal_svp = 0;
 				}
 				else if (strnicmp(cmdbuffer, rbkcmd, strlen(rbkcmd)) == 0)
 				{
 					CC_mark_cursors_doubtful(self);
 					CC_set_in_error_trans(self); /* mark the transaction error in case of manual rollback */
+					self->internal_svp = 0; /* possibly an internal savepoint is invalid */
+					self->opt_previous = 0; /* unknown */
+					CC_init_opt_in_progress(self);
+				}
+				else if (strnicmp(cmdbuffer, rlscmd, strlen(rlscmd)) == 0)
+				{
+					if (discard_next_release)
+					{
+inolog("Discarded a RELEASE result\n");
+						discard_next_release = FALSE;
+						break; /* discard the result */
+					}
+					self->internal_svp = 0;
+					if (SAVEPOINT_IN_PROGRESS == self->internal_op)
+						break; /* discard the result */
 				}
 				/*
 				 *	DROP TABLE or ALTER TABLE may change
@@ -2036,19 +2142,13 @@ cleanup:
 		PQclear(pgres);
 		pgres = NULL;
 	}
+inolog(" !!!! %s:rollback_on_error=%d CC_is_in_trans=%d discard_next_savepoint=%d query_rollback=%d\n", __FUNCTION__, rollback_on_error, CC_is_in_trans(self), discard_next_savepoint, query_rollback);
 	if (rollback_on_error && CC_is_in_trans(self) && !discard_next_savepoint)
 	{
 		if (query_rollback)
 		{
-			if (CC_is_in_error_trans(self))
-			{
-				char tmpsqlbuf[100];
-				snprintf(tmpsqlbuf, sizeof(tmpsqlbuf),
-						 "%s TO %s; %s %s",
-						 rbkcmd, per_query_svp,
-						 rlscmd, per_query_svp);
-				pgres = PQexec(self->pqconn, tmpsqlbuf);
-			}
+			if (!CC_internal_rollback(self, PER_QUERY_ROLLBACK, ignore_abort_on_conn))
+				ignore_abort_on_conn = FALSE;
 		}
 		else if (CC_is_in_error_trans(self))
 			pgres = PQexec(self->pqconn, rbkcmd);
@@ -2113,14 +2213,19 @@ cleanup:
 				 *	If error message isn't set
 				 */
 				if (ignore_abort_on_conn)
-					CC_set_errornumber(self, 0);
+				{
+					CC_set_errornumber(self, CONN_ERROR_IGNORED);
+					if (retres)
+						QR_set_rstatus(retres, PORES_NONFATAL_ERROR);
+inolog(" !!!! %s:ignored abort_on_conn\n", __FUNCTION__);
+				}
 				else if (retres)
 				{
 					if (NULL == CC_get_errormsg(self) ||
 					    !CC_get_errormsg(self)[0])
 						CC_set_errormsg(self, QR_get_message(retres));
 					if (!self->sqlstate[0])
-						strcpy(self->sqlstate, retres->sqlstate);
+						STRCPY_FIXED(self->sqlstate, retres->sqlstate);
 				}
 			}
 		}
@@ -2202,7 +2307,7 @@ CC_send_function(ConnectionClass *self, const char *fn_name, void *result_buf, i
 #define	return DONT_CALL_RETURN_FROM_HERE???
 	ENTER_INNER_CONN_CS(self, func_cs_count);
 
-	snprintf(sqlbuffer, sizeof(sqlbuffer), "SELECT pg_catalog.%s%s", fn_name,
+	SPRINTF_FIXED(sqlbuffer, "SELECT pg_catalog.%s%s", fn_name,
 			 func_param_str[nargs]);
 	for (i = 0; i < nargs; ++i)
 	{
@@ -2292,7 +2397,6 @@ char
 CC_send_settings(ConnectionClass *self, const char *set_query)
 {
 	HSTMT		hstmt;
-	StatementClass *stmt;
 	RETCODE		result;
 	char		status = TRUE;
 	char	   *cs,
@@ -2315,9 +2419,8 @@ CC_send_settings(ConnectionClass *self, const char *set_query)
 	result = PGAPI_AllocStmt(self, &hstmt, 0);
 	if (!SQL_SUCCEEDED(result))
 		return FALSE;
-	stmt = (StatementClass *) hstmt;
 
-	stmt->internal = TRUE;		/* ensure no BEGIN/COMMIT/ABORT stuff */
+	/* non-external handle ensures no BEGIN/COMMIT/ABORT stuff */
 
 	cs = strdup(set_query);
 	if (cs == NULL)
@@ -2367,7 +2470,7 @@ CC_lookup_lo(ConnectionClass *self)
 	mylog("%s: entering...\n", func);
 
 	res = CC_send_query(self, "select oid, typbasetype from pg_type where typname = '"  PG_TYPE_LO_NAME "'",
-		NULL, IGNORE_ABORT_ON_CONN | ROLLBACK_ON_ERROR, NULL);
+		NULL, READ_ONLY_QUERY, NULL);
 	if (QR_command_maybe_successful(res) && QR_get_num_cached_tuples(res) > 0)
 	{
 		OID	basetype;
@@ -2394,7 +2497,7 @@ CC_lookup_lo(ConnectionClass *self)
 void
 CC_initialize_pg_version(ConnectionClass *self)
 {
-	strcpy(self->pg_version, "7.4");
+	STRCPY_FIXED(self->pg_version, "7.4");
 	self->pg_version_major = 7;
 	self->pg_version_minor = 4;
 }
@@ -2431,7 +2534,7 @@ CC_get_current_schema(ConnectionClass *conn)
 	{
 		QResultClass	*res;
 
-		if (res = CC_send_query(conn, "select current_schema()", NULL, IGNORE_ABORT_ON_CONN | ROLLBACK_ON_ERROR, NULL), QR_command_maybe_successful(res))
+		if (res = CC_send_query(conn, "select current_schema()", NULL, READ_ONLY_QUERY, NULL), QR_command_maybe_successful(res))
 		{
 			if (QR_get_num_total_tuples(res) == 1)
 			{
@@ -2449,15 +2552,16 @@ CC_get_current_schema(ConnectionClass *conn)
 
 int	CC_mark_a_object_to_discard(ConnectionClass *conn, int type, const char *plan)
 {
-	int	cnt = conn->num_discardp + 1;
+	int	cnt = conn->num_discardp + 1, plansize;
 	char	*pname;
 
 	CC_REALLOC_return_with_error(conn->discardp, char *,
 		(cnt * sizeof(char *)), conn, "Couldn't alloc discardp.", -1);
-	CC_MALLOC_return_with_error(pname, char, (strlen(plan) + 2),
+	plansize = strlen(plan) + 2;
+	CC_MALLOC_return_with_error(pname, char, plansize,
 		conn, "Couldn't alloc discardp mem.", -1);
 	pname[0] = (char) type;	/* 's':prepared statement 'p':cursor */
-	strcpy(pname + 1, plan);
+	strncpy_null(pname + 1, plan, plansize - 1);
 	conn->discardp[conn->num_discardp++] = pname;
 
 	return 1;
@@ -2475,10 +2579,10 @@ int	CC_discard_marked_objects(ConnectionClass *conn)
 	{
 		pname = conn->discardp[i];
 		if ('s' == pname[0])
-			snprintf(cmd, sizeof(cmd), "DEALLOCATE \"%s\"", pname + 1);
+			SPRINTF_FIXED(cmd, "DEALLOCATE \"%s\"", pname + 1);
 		else
-			snprintf(cmd, sizeof(cmd), "CLOSE \"%s\"", pname + 1);
-		res = CC_send_query(conn, cmd, NULL, ROLLBACK_ON_ERROR | IGNORE_ABORT_ON_CONN, NULL);
+			SPRINTF_FIXED(cmd, "CLOSE \"%s\"", pname + 1);
+		res = CC_send_query(conn, cmd, NULL, ROLLBACK_ON_ERROR | IGNORE_ABORT_ON_CONN | READ_ONLY_QUERY, NULL);
 		QR_Destructor(res);
 		free(conn->discardp[i]);
 		conn->num_discardp--;
@@ -2558,9 +2662,9 @@ LIBPQ_connect(ConnectionClass *self)
 		char emsg[200];
 
 		if (errmsg != NULL)
-			snprintf(emsg, sizeof(emsg), "libpq connection parameter error:%s", errmsg);
+			SPRINTF_FIXED(emsg, "libpq connection parameter error:%s", errmsg);
 		else
-			strncpy_null(emsg, "memory error? in PQconninfoParse", sizeof(emsg));
+			STRCPY_FIXED(emsg, "memory error? in PQconninfoParse");
 		CC_set_error(self, CONN_OPENDB_ERROR, emsg, func);
 		goto cleanup;
 	}
@@ -2618,17 +2722,17 @@ LIBPQ_connect(ConnectionClass *self)
 	}
 	if (self->login_timeout > 0)
 	{
-		sprintf(login_timeout_str, "%u", (unsigned int) self->login_timeout);
+		SPRINTF_FIXED(login_timeout_str, "%u", (unsigned int) self->login_timeout);
 		opts[cnt] = "connect_timeout";	vals[cnt++] = login_timeout_str;
 	}
 	if (self->connInfo.keepalive_idle > 0)
 	{
-		sprintf(keepalive_idle_str, "%d", self->connInfo.keepalive_idle);
+		ITOA_FIXED(keepalive_idle_str, self->connInfo.keepalive_idle);
 		opts[cnt] = "keepalives_idle";	vals[cnt++] = keepalive_idle_str;
 	}
 	if (self->connInfo.keepalive_interval > 0)
 	{
-		sprintf(keepalive_interval_str, "%d", self->connInfo.keepalive_interval);
+		ITOA_FIXED(keepalive_interval_str, self->connInfo.keepalive_interval);
 		opts[cnt] = "keepalives_interval";	vals[cnt++] = keepalive_interval_str;
 	}
 	if (conninfoOption != NULL)
@@ -2648,7 +2752,7 @@ LIBPQ_connect(ConnectionClass *self)
 
 						if (vals[j] != NULL && strcmp(vals[j], val) == 0)
 							continue;
-						snprintf(emsg, sizeof(emsg), "%s parameter in pqopt option conflicts with other ordinary option", keyword);
+						SPRINTF_FIXED(emsg, "%s parameter in pqopt option conflicts with other ordinary option", keyword);
 						CC_set_error(self, CONN_OPENDB_ERROR, emsg, func);
 						goto cleanup;
 					}
@@ -2708,14 +2812,14 @@ inolog("status=%d\n", pqret);
 	pversion = PQserverVersion(pqconn);
 	self->pg_version_major = pversion / 10000;
 	self->pg_version_minor = (pversion % 10000) / 100;
-	sprintf(self->pg_version, "%d.%d.%d",  self->pg_version_major, self->pg_version_minor, pversion % 100);
+	SPRINTF_FIXED(self->pg_version, "%d.%d.%d",  self->pg_version_major, self->pg_version_minor, pversion % 100);
 
 	mylog("Server version=%s\n", self->pg_version);
 
 	if (!CC_get_username(self)[0])
 	{
 		mylog("PQuser=%s\n", PQuser(pqconn));
-		strncpy_null(self->connInfo.username, PQuser(pqconn), sizeof(self->connInfo.username));
+		STRCPY_FIXED(self->connInfo.username, PQuser(pqconn));
 	}
 
 	ret = 1;
@@ -2836,14 +2940,13 @@ make_lstring_ifneeded(ConnectionClass *conn, const SQLCHAR *s, ssize_t len, BOOL
  *	This routine could be modified to use vsprintf() to handle multiple arguments.
  */
 static char *
-my_strcat(char *buf, const char *fmt, const char *s, ssize_t len)
+my_strcat(char *buf, int buflen, const char *fmt, const char *s, ssize_t len)
 {
-	if (s && (len > 0 || (len == SQL_NTS && strlen(s) > 0)))
+	if (s && (len > 0 || (len == SQL_NTS && *s != 0)))
 	{
-		size_t			length = (len > 0) ? len : strlen(s);
-		size_t			pos = strlen(buf);
+		size_t	length = (len > 0) ? len : strlen(s);
 
-		sprintf(&buf[pos], fmt, length, s);
+		snprintfcat(buf, buflen, fmt, length, s);
 		return buf;
 	}
 	return NULL;
@@ -2854,24 +2957,23 @@ my_strcat(char *buf, const char *fmt, const char *s, ssize_t len)
  *	It can have 1 more parameter than my_strcat.
  */
 static char *
-my_strcat1(char *buf, const char *fmt, const char *s1, const char *s)
+my_strcat1(char *buf, int buflen, const char *fmt, const char *s1, const char *s)
 {
 	if (s && s[0] != '\0')
 	{
-		size_t	pos = strlen(buf);
 		ssize_t	length = strlen(s);
 
 		if (s1)
-			sprintf(&buf[pos], fmt, s1, length, s);
+			snprintfcat(buf, buflen, fmt, s1, length, s);
 		else
-			sprintf(&buf[pos], fmt, length, s);
+			snprintfcat(buf, buflen, fmt, length, s);
 		return buf;
 	}
 	return NULL;
 }
 
 char *
-schema_strcat(char *buf, const char *fmt, const SQLCHAR *s, SQLLEN len, const SQLCHAR *tbname, SQLLEN tbnmlen, ConnectionClass *conn)
+schema_strcat(char *buf, int buflen, const char *fmt, const SQLCHAR *s, SQLLEN len, const SQLCHAR *tbname, SQLLEN tbnmlen, ConnectionClass *conn)
 {
 	if (!s || 0 == len)
 	{
@@ -2881,22 +2983,22 @@ schema_strcat(char *buf, const char *fmt, const SQLCHAR *s, SQLLEN len, const SQ
 		 * naming.
 		 */
 		if (tbname && (tbnmlen > 0 || tbnmlen == SQL_NTS))
-			return my_strcat(buf, fmt, CC_get_current_schema(conn), SQL_NTS);
+			return my_strcat(buf, buflen, fmt, CC_get_current_schema(conn), SQL_NTS);
 		return NULL;
 	}
-	return my_strcat(buf, fmt, (char *) s, len);
+	return my_strcat(buf, buflen, fmt, (char *) s, len);
 }
 
 char *
-schema_strcat1(char *buf, const char *fmt, const char *s1, const char *s, const SQLCHAR *tbname, int tbnmlen, ConnectionClass *conn)
+schema_strcat1(char *buf, int buflen, const char *fmt, const char *s1, const char *s, const SQLCHAR *tbname, int tbnmlen, ConnectionClass *conn)
 {
 	if (!s || s[0] == '\0')
 	{
 		if (tbname && (tbnmlen > 0 || tbnmlen == SQL_NTS))
-			return my_strcat1(buf, fmt, s1, CC_get_current_schema(conn));
+			return my_strcat1(buf, buflen, fmt, s1, CC_get_current_schema(conn));
 		return NULL;
 	}
-	return my_strcat1(buf, fmt, s1, s);
+	return my_strcat1(buf, buflen, fmt, s1, s);
 }
 
 #ifdef	_HANDLE_ENLIST_IN_DTC_
@@ -2952,6 +3054,7 @@ DLL_DECLARE void PgDtc_create_connect_string(void *self, char *connstr, int strs
 		 * for the XA transaction manager, but pick the same as the client driver
 		 * to keep things as similar as possible.
 		 */
+		if (strcmp(drivername, DBMS_NAME) == 0)
 #ifdef UNICODE_SUPPORT
 		drivername = DBMS_NAME_UNICODE"(x64)";
 #else
@@ -2965,10 +3068,10 @@ DLL_DECLARE void PgDtc_create_connect_string(void *self, char *connstr, int strs
 	{
 		case DTC_CHECK_LINK_ONLY:
 		case DTC_CHECK_BEFORE_LINK:
-			sprintf(xaOptStr, KEYWORD_DTC_CHECK "=0;");
+			SPRINTF_FIXED(xaOptStr, KEYWORD_DTC_CHECK "=0;");
 			break;
 		case DTC_CHECK_RM_CONNECTION:
-			sprintf(xaOptStr, KEYWORD_DTC_CHECK "=1;");
+			SPRINTF_FIXED(xaOptStr, KEYWORD_DTC_CHECK "=1;");
 			break;
 		default:
 			*xaOptStr = '\0';
@@ -3219,13 +3322,13 @@ PgDtc_two_phase_operation(void *self, int operation, const char *gxid)
 	switch (operation)
 	{
 		case PREPARE_TRANSACTION:
-			snprintf(cmd, sizeof(cmd), "PREPARE TRANSACTION '%s'", gxid);
+			SPRINTF_FIXED(cmd, "PREPARE TRANSACTION '%s'", gxid);
 			break;
 		case COMMIT_PREPARED:
-			snprintf(cmd, sizeof(cmd), "COMMIT PREPARED '%s'", gxid);
+			SPRINTF_FIXED(cmd, "COMMIT PREPARED '%s'", gxid);
 			break;
 		case ROLLBACK_PREPARED:
-			snprintf(cmd, sizeof(cmd), "ROLLBACK PREPARED '%s'", gxid);
+			SPRINTF_FIXED(cmd, "ROLLBACK PREPARED '%s'", gxid);
 			break;
 	}
 
@@ -3339,9 +3442,9 @@ CC_set_transact(ConnectionClass *self, UInt4 isolation)
 	if (self->default_isolation == 0)
 		bShow = TRUE;
 	if (bShow)
-		res = CC_send_query_append(self, ISOLATION_SHOW_QUERY, NULL, 0, NULL, query);
+		res = CC_send_query_append(self, ISOLATION_SHOW_QUERY, NULL, READ_ONLY_QUERY, NULL, query);
 	else
-		res = CC_send_query(self, query, NULL, 0, NULL);
+		res = CC_send_query(self, query, NULL, READ_ONLY_QUERY, NULL);
 	if (!QR_command_maybe_successful(res))
 	{
 		CC_set_error(self, CONN_EXEC_ERROR, "ISOLATION change request to the server error", __FUNCTION__);
