@@ -35,9 +35,6 @@
 
 static const char *next_name_token(const char *s, size_t *len);
 
-/*extern GLOBAL_VALUES globals;*/
-
-
 /*		Perform a Prepare on the SQL statement */
 RETCODE		SQL_API
 PGAPI_Prepare(HSTMT hstmt,
@@ -134,8 +131,6 @@ PGAPI_Prepare(HSTMT hstmt,
 cleanup:
 #undef	return
 inolog("SQLPrepare return=%d\n", retval);
-	if (self->internal)
-		retval = DiscardStatementSvp(self, retval, FALSE);
 	return retval;
 }
 
@@ -173,6 +168,8 @@ inolog("a2\n");
 
 	if (0 != (flag & PODBC_WITH_HOLD))
 		SC_set_with_hold(stmt);
+	if (0 != (flag & PODBC_RDONLY))
+		SC_set_readonly(stmt);
 
 	/*
 	 * If an SQLPrepare was performed prior to this, but was left in the
@@ -398,6 +395,13 @@ int HowToPrepareBeforeExec(StatementClass *stmt, BOOL checkOnly)
 	return nCallParse;
 }
 
+static
+const char *GetSvpName(const ConnectionClass *conn, char *wrk, int wrksize)
+{
+	snprintf(wrk, wrksize, "_EXEC_SVP_%p", conn);
+	return wrk;
+}
+
 /*
  *	The execution after all parameters were resolved.
  */
@@ -550,7 +554,7 @@ StartRollbackState(StatementClass *stmt)
 	ConnectionClass	*conn;
 	ConnInfo	*ci = NULL;
 
-inolog("%s:%p->internal=%d\n", func, stmt, stmt->internal);
+inolog("%s:%p->external=%d\n", func, stmt, stmt->external);
 	conn = SC_get_conn(stmt);
 	if (conn)
 		ci = &conn->connInfo;
@@ -581,15 +585,43 @@ inolog("%s:%p->internal=%d\n", func, stmt, stmt->internal);
 	return	ret;
 }
 
+int
+GenerateSvpCommand(ConnectionClass *conn, int type, char *cmd, int buflen)
+{
+	char esavepoint[32];
+	int	rtn = -1;
+
+	cmd[0] = '\0';
+	switch (type)
+	{
+		case INTERNAL_SAVEPOINT_OPERATION:	/* savepoint */
+#ifdef	_RELEASE_INTERNAL_SAVEPOINT
+			if (conn->internal_svp)
+				rtn = snprintf(cmd, buflen, "RELEASE %s;", GetSvpName(conn, esavepoint, sizeof(esavepoint)));
+#endif /* _RELEASE_INTERNAL_SAVEPOINT */
+			rtn = snprintfcat(cmd, buflen, "SAVEPOINT %s", GetSvpName(conn, esavepoint, sizeof(esavepoint)));
+			break;
+		case INTERNAL_ROLLBACK_OPERATION: /* rollback */
+			if (conn->internal_svp)
+				rtn = snprintf(cmd, buflen, "ROLLBACK TO %s", GetSvpName(conn, esavepoint, sizeof(esavepoint)));
+			else
+				rtn = snprintf(cmd, buflen, "ROLLBACK");
+
+			break;
+	}
+
+	return rtn;
+}
+
 /*
  *	Must be in a transaction or the subsequent execution
  *	invokes a transaction.
  */
 RETCODE
-SetStatementSvp(StatementClass *stmt)
+SetStatementSvp(StatementClass *stmt, unsigned int option)
 {
 	CSTR	func = "SetStatementSvp";
-	char	esavepoint[32], cmd[64];
+	char	cmd[64];
 	ConnectionClass	*conn = SC_get_conn(stmt);
 	QResultClass *res;
 	RETCODE	ret = SQL_SUCCESS_WITH_INFO;
@@ -602,46 +634,45 @@ SetStatementSvp(StatementClass *stmt)
 	if (CC_is_in_error_trans(conn))
 		return ret;
 
-	if (0 == stmt->lock_CC_for_rb)
+	if (0 == conn->lock_CC_for_rb)
 	{
 		ENTER_CONN_CS(conn);
-		stmt->lock_CC_for_rb++;
+		conn->lock_CC_for_rb++;
 	}
+inolog(" !!!! %s:%p->accessed=%d opt=%u in_progress=%u prev=%u\n", __FUNCTION__, conn, CC_accessed_db(conn), option, conn->opt_in_progress, conn->opt_previous);
+	conn->opt_in_progress &= option;
 	switch (stmt->statement_type)
 	{
 		case STMT_TYPE_SPECIAL:
 		case STMT_TYPE_TRANSACTION:
 			return ret;
 	}
-	if (!SC_accessed_db(stmt))
+	/* If rbpoint is not yet started and the previous statement was not read-only */
+	if (!CC_started_rbpoint(conn) && 0 == (conn->opt_previous & SVPOPT_RDONLY))
 	{
 		BOOL	need_savep = FALSE;
 
-		if (stmt->internal)
-		{
-			if (PG_VERSION_GE(conn, 8.0))
-				SC_start_rb_stmt(stmt);
-			else
-				SC_start_tc_stmt(stmt);
-		}
 		if (SC_is_rb_stmt(stmt))
 		{
-			if (CC_is_in_trans(conn))
+			if (CC_is_in_trans(conn)) /* needless to issue SAVEPOINT before the 1st command */
 			{
 				need_savep = TRUE;
 			}
 		}
 		if (need_savep)
 		{
-			sprintf(esavepoint, "_EXEC_SVP_%p", stmt);
-			snprintf(cmd, sizeof(cmd), "SAVEPOINT %s", esavepoint);
-			res = CC_send_query(conn, cmd, NULL, 0, NULL);
-			if (QR_command_maybe_successful(res))
+			if (0 != (option & SVPOPT_REDUCE_ROUNDTRIP))
 			{
-				SC_set_accessed_db(stmt);
-				SC_start_rbpoint(stmt);
-				ret = SQL_SUCCESS;
+				conn->internal_op = PREPEND_IN_PROGRESS;
+				CC_set_accessed_db(conn);
+				return ret;
 			}
+			GenerateSvpCommand(conn, INTERNAL_SAVEPOINT_OPERATION, cmd, sizeof(cmd));
+			conn->internal_op = SAVEPOINT_IN_PROGRESS;
+			res = CC_send_query(conn, cmd, NULL, 0, NULL);
+			conn->internal_op = 0;
+			if (QR_command_maybe_successful(res))
+				ret = SQL_SUCCESS;
 			else
 			{
 				SC_set_error(stmt, STMT_INTERNAL_ERROR, "internal SAVEPOINT failed", func);
@@ -649,10 +680,9 @@ SetStatementSvp(StatementClass *stmt)
 			}
 			QR_Destructor(res);
 		}
-		else
-			SC_set_accessed_db(stmt);
 	}
-inolog("%s:%p->accessed=%d\n", func, stmt, SC_accessed_db(stmt));
+	CC_set_accessed_db(conn);
+inolog("%s:%p->accessed=%d\n", func, conn, CC_accessed_db(conn));
 	return ret;
 }
 
@@ -660,13 +690,13 @@ RETCODE
 DiscardStatementSvp(StatementClass *stmt, RETCODE ret, BOOL errorOnly)
 {
 	CSTR	func = "DiscardStatementSvp";
-	char	esavepoint[32], cmd[64];
 	ConnectionClass	*conn = SC_get_conn(stmt);
-	QResultClass *res;
-	BOOL	cmd_success, start_stmt = FALSE;
+	BOOL	start_stmt = FALSE;
 
-inolog("%s:%p->accessed=%d is_in=%d is_rb=%d is_tc=%d\n", func, stmt, SC_accessed_db(stmt),
+inolog("%s:%p->accessed=%d is_in=%d is_rb=%d is_tc=%d\n", func, conn, CC_accessed_db(conn),
 CC_is_in_trans(conn), SC_is_rb_stmt(stmt), SC_is_tc_stmt(stmt));
+	if (conn->lock_CC_for_rb > 0)
+		mylog("%s:in_progress=%u previous=%d\n", func, conn->opt_in_progress, conn->opt_previous);
 	switch (ret)
 	{
 		case SQL_NEED_DATA:
@@ -679,23 +709,19 @@ CC_is_in_trans(conn), SC_is_rb_stmt(stmt), SC_is_tc_stmt(stmt));
 				start_stmt = TRUE;
 			break;
 	}
-	if (!SC_accessed_db(stmt) || !CC_is_in_trans(conn))
+	if (!CC_accessed_db(conn) || !CC_is_in_trans(conn))
 		goto cleanup;
 	if (!SC_is_rb_stmt(stmt) && !SC_is_tc_stmt(stmt))
 		goto cleanup;
-	sprintf(esavepoint, "_EXEC_SVP_%p", stmt);
 	if (SQL_ERROR == ret)
 	{
-		if (SC_started_rbpoint(stmt))
+		if (CC_started_rbpoint(conn) && conn->internal_svp)
 		{
-			snprintf(cmd, sizeof(cmd), "ROLLBACK to %s", esavepoint);
-			res = CC_send_query(conn, cmd, NULL, IGNORE_ABORT_ON_CONN, NULL);
-			cmd_success = QR_command_maybe_successful(res);
-			QR_Destructor(res);
+			int	cmd_success = CC_internal_rollback(conn, PER_STATEMENT_ROLLBACK, FALSE);
+
 			if (!cmd_success)
 			{
 				SC_set_error(stmt, STMT_INTERNAL_ERROR, "internal ROLLBACK failed", func);
-				CC_abort(conn);
 				goto cleanup;
 			}
 		}
@@ -708,19 +734,6 @@ CC_is_in_trans(conn), SC_is_rb_stmt(stmt), SC_is_tc_stmt(stmt));
 	else if (errorOnly)
 		return ret;
 inolog("ret=%d\n", ret);
-	if (SQL_NEED_DATA != ret && SC_started_rbpoint(stmt))
-	{
-		snprintf(cmd, sizeof(cmd), "RELEASE %s", esavepoint);
-		res = CC_send_query(conn, cmd, NULL, IGNORE_ABORT_ON_CONN, NULL);
-		cmd_success = QR_command_maybe_successful(res);
-		QR_Destructor(res);
-		if (!cmd_success)
-		{
-			SC_set_error(stmt, STMT_INTERNAL_ERROR, "internal RELEASE failed", func);
-			CC_abort(conn);
-			ret = SQL_ERROR;
-		}
-	}
 cleanup:
 #ifdef NOT_USED
 	if (!SC_is_prepare_statement(stmt) && ONCE_DESCRIBED == stmt->prepared)
@@ -728,12 +741,19 @@ cleanup:
 #endif
 	if (start_stmt || SQL_ERROR == ret)
 	{
-		if (stmt->lock_CC_for_rb > 0)
+		stmt->execinfo = 0;
+		if (SQL_ERROR != ret && CC_accessed_db(conn))
+		{
+			conn->opt_previous = conn->opt_in_progress;
+			CC_init_opt_in_progress(conn);
+		}
+		while (conn->lock_CC_for_rb > 0)
 		{
 			LEAVE_CONN_CS(conn);
-			stmt->lock_CC_for_rb--;
+			conn->lock_CC_for_rb--;
+			inolog(" %s:release conn_lock\n", __FUNCTION__);
 		}
-		SC_start_stmt(stmt);
+		CC_start_stmt(conn);
 	}
 	return ret;
 }
@@ -988,7 +1008,7 @@ mylog("prepareParameters was %s called, prepare state:%d\n", shouldParse == nCal
 		if (recycle && !recycled)
 			SC_recycle_statement(stmt);
 		if (isSqlServr() &&
-		    !stmt->internal &&
+		    stmt->external &&
 		    0 != stmt->prepare &&
 		    PG_VERSION_LT(conn, 8.4) &&
 		    SC_can_parse_statement(stmt))
@@ -1089,8 +1109,6 @@ cleanup:
 mylog("retval=%d\n", retval);
 	SC_setInsertedTable(stmt, retval);
 #undef	return
-	if (stmt->internal)
-		retval = DiscardStatementSvp(stmt, retval, FALSE);
 	return retval;
 }
 
@@ -1220,8 +1238,6 @@ PGAPI_Cancel(HSTMT hstmt)		/* Statement to cancel. */
 		estmt->current_exec_param = -1;
 		estmt->put_data = FALSE;
 		cancelNeedDataState(estmt);
-		if (stmt->internal)
-			ret = DiscardStatementSvp(stmt, ret, FALSE);
 		LEAVE_STMT_CS(stmt);
 		return ret;
 	}
@@ -1435,8 +1451,6 @@ inolog("return SQL_NEED_DATA\n");
 cleanup:
 #undef	return
 	SC_setInsertedTable(stmt, retval);
-	if (stmt->internal)
-		retval = DiscardStatementSvp(stmt, retval, FALSE);
 	mylog("%s: returning %d\n", func, retval);
 	return retval;
 }
@@ -1679,8 +1693,6 @@ cleanup:
 #undef	return
 	if (allocbuf)
 		free(allocbuf);
-	if (stmt->internal)
-		retval = DiscardStatementSvp(stmt, retval, TRUE);
 
 	return retval;
 }
