@@ -158,12 +158,12 @@ my_trim(char *s)
 }
 
 /*
- * snprintf_add is a extension to snprintf
+ * snprintfcat is a extension to snprintf
  * It add format to buf at given pos
  */
 
 int
-snprintf_add(char *buf, size_t size, const char *format, ...)
+snprintfcat(char *buf, size_t size, const char *format, ...)
 {
 	int len;
 	size_t pos = strlen(buf);
@@ -172,11 +172,11 @@ snprintf_add(char *buf, size_t size, const char *format, ...)
 	va_start(arglist, format);
 	len = vsnprintf(buf + pos, size - pos, format, arglist);
 	va_end(arglist);
-	return len;
+	return len + pos;
 }
 
 /*
- * snprintf_addlen is a extension to snprintf
+ * snprintf_len is a extension to snprintf
  * It returns strlen of buf every time (not -1 when truncated)
  */
 
@@ -203,6 +203,7 @@ snprintf_len(char *buf, size_t size, const char *format, ...)
  * and repeat.
  */
 #ifdef WIN32
+#if (_MSC_VER < 1900)	/* vc14 provieds snprintf() */
 static int
 posix_vsnprintf(char *str, size_t size, const char *format, va_list ap)
 {
@@ -248,7 +249,8 @@ posix_snprintf(char *buf, size_t size, const char *format, ...)
 	va_end(arglist);
 	return len;
 }
-#endif
+#endif /* _MSC_VER */
+#endif /* WIN32 */
 
 #ifndef	HAVE_STRLCAT
 size_t
