@@ -27,6 +27,7 @@
 #include "qresult.h"
 #include "pgapifunc.h"
 #include "loadlib.h"
+#include "dlg_specific.h"
 
 
 /*	SQLError -> SQLDiagRec */
@@ -1743,6 +1744,7 @@ PGAPI_SetConnectAttr(HDBC ConnectionHandle,
 			break;
 		case SQL_ATTR_PGOPT_USE_DECLAREFETCH:
 			conn->connInfo.drivers.use_declarefetch = CAST_UPTR(SQLCHAR, Value);
+			ci_updatable_cursors_set(&conn->connInfo);
 			qlog("declarefetch => %d\n", conn->connInfo.drivers.use_declarefetch);
 			mylog("declarefetch => %d\n", conn->connInfo.drivers.use_declarefetch);
 			break;
@@ -1800,7 +1802,7 @@ PGAPI_SetConnectAttr(HDBC ConnectionHandle,
 	if (unsupported)
 	{
 		char	msg[64];
-		snprintf(msg, sizeof(msg), "Couldn't set unsupported connect attribute " FORMAT_INTEGER, Attribute);
+		SPRINTF_FIXED(msg, "Couldn't set unsupported connect attribute " FORMAT_INTEGER, Attribute);
 		CC_set_error(conn, CONN_OPTION_NOT_FOR_THE_DRIVER, msg, func);
 		return SQL_ERROR;
 	}
@@ -2028,7 +2030,6 @@ RETCODE	bulk_ope_callback(RETCODE retcode, void *para)
 	ConnectionClass	*conn;
 	QResultClass	*res;
 	IRDFields	*irdflds;
-	BindInfoClass	*bookmark;
 	PG_BM		pg_bm;
 
 	if (s->need_data_callback)
@@ -2042,7 +2043,6 @@ RETCODE	bulk_ope_callback(RETCODE retcode, void *para)
 		s->idx = s->processed = 0;
 	}
 	s->need_data_callback = FALSE;
-	bookmark = s->opts->bookmark;
 	res = SC_get_Curres(s->stmt);
 	for (; SQL_ERROR != ret && s->idx < s->opts->size_of_rowset; s->idx++)
 	{
@@ -2132,7 +2132,5 @@ PGAPI_BulkOperations(HSTMT hstmt, SQLSMALLINT operationX)
 		s.need_data_callback = FALSE;
 		ret = bulk_ope_callback(SQL_SUCCESS, &s);
 	}
-	if (s.stmt->internal)
-		ret = DiscardStatementSvp(s.stmt, ret, FALSE);
 	return ret;
 }
