@@ -232,15 +232,17 @@ typedef double SDOUBLE;
 #define	FALSE	(BOOL)0
 #endif /* FALSE */
 #else
+#if (_MSC_VER < 1900) /* vc12 or under */
 #define snprintf posix_snprintf
 extern int posix_snprintf(char *buf, size_t size, const char *format, ...);
+#define vsnprintf _vsnprintf
+#endif /* _MSC_VER */
 
 #ifndef strdup
 #define strdup _strdup
 #endif /* strdup */
 #define strnicmp _strnicmp
 #define stricmp _stricmp
-#define vsnprintf _vsnprintf
 #endif /* WIN32 */
 
 #ifndef	SQL_ATTR_APP_ROW_DESC
@@ -265,7 +267,10 @@ extern int posix_snprintf(char *buf, size_t size, const char *format, ...);
 
 #define DRIVER_ODBC_VER				"03.51"
 
-#ifndef	UNICODE_SUPPORT
+#ifdef	UNICODE_SUPPORT
+#define WCLEN	sizeof(SQLWCHAR)
+SQLULEN ucs2strlen(const SQLWCHAR *);
+#else
 #undef	SQL_WCHAR
 #undef	SQL_WVARCHAR
 #undef	SQL_WLONGVARCHAR
@@ -275,12 +280,21 @@ extern int posix_snprintf(char *buf, size_t size, const char *format, ...);
 #define	SQL_WLONGVARCHAR	SQL_WLONGVARCHAR_IS_INHIBITED
 #define	SQL_C_WCHAR	SQL_C_WCHAR_IS_INHIBITED
 #endif
+
 #ifndef DBMS_NAME
+#ifdef	_WIN64
+#ifdef	UNICODE_SUPPORT
+#define DBMS_NAME				DBMS_NAME_UNICODE"(x64)"
+#else
+#define DBMS_NAME				DBMS_NAME_ANSI"(x64)"
+#endif /* UNICODE_SUPPORT */
+#else	/* _WIN64 */
 #ifdef	UNICODE_SUPPORT
 #define DBMS_NAME				DBMS_NAME_UNICODE
 #else
 #define DBMS_NAME				DBMS_NAME_ANSI
 #endif /* UNICODE_SUPPORT */
+#endif /* _WIN64 */
 #endif /* DBMS_NAME */
 
 #ifndef DBMS_NAME
@@ -290,7 +304,7 @@ extern int posix_snprintf(char *buf, size_t size, const char *format, ...);
 #ifdef	UNICODE_SUPPORT
 #define DRIVER_FILE_NAME			"PSQLODBC35W.DLL"
 #else
-#define DRIVER_FILE_NAME			"PSQLODBC.DLL"
+#define DRIVER_FILE_NAME			"PSQLODBC30A.DLL"
 #endif   /* UNICODE_SUPPORT */
 #else
 #ifdef  UNICODE_SUPPORT
@@ -452,6 +466,7 @@ do { \
 #define	NAMEICMP(name1, name2) (stricmp(SAFE_NAME(name1), SAFE_NAME(name2)))
 /*	pgNAME define end */
 
+
 typedef struct GlobalValues_
 {
 	pgNAME		drivername;
@@ -462,8 +477,6 @@ typedef struct GlobalValues_
 	char		debug;
 	char		commlog;
 	char		unique_index;
-	char		onlyread;		/* readonly is reserved on Digital C++
-								 * compiler */
 	char		use_declarefetch;
 	char		text_as_longvarchar;
 	char		unknowns_as_longvarchar;
@@ -472,10 +485,10 @@ typedef struct GlobalValues_
 	char		parse;
 	char		extra_systable_prefixes[MEDIUM_REGISTRY_LEN];
 	char		protocol[SMALL_REGISTRY_LEN];
-	pgNAME		conn_settings;
 } GLOBAL_VALUES;
 
 void copy_globals(GLOBAL_VALUES *to, const GLOBAL_VALUES *from);
+void init_globals(GLOBAL_VALUES *glbv);
 void finalize_globals(GLOBAL_VALUES *glbv);
 
 typedef struct StatementOptions_
@@ -505,17 +518,17 @@ typedef struct QueryInfo_
 typedef struct
 {
         UInt4	status;
-        Int4	errorsize;
+        Int2	errorsize;
         Int2	recsize;
         Int2	errorpos;
-        char    sqlstate[8];
+        char    sqlstate[6];
         SQLLEN	diag_row_count;
-        char    __error_message[1];
+        char    __error_message[40];
 }       PG_ErrorInfo;
 PG_ErrorInfo	*ER_Constructor(SDWORD errornumber, const char *errormsg);
 PG_ErrorInfo	*ER_Dup(const PG_ErrorInfo *from);
 void ER_Destructor(PG_ErrorInfo *);
-RETCODE SQL_API ER_ReturnError(PG_ErrorInfo **, SQLSMALLINT, UCHAR *,
+RETCODE SQL_API ER_ReturnError(PG_ErrorInfo *, SQLSMALLINT, UCHAR *,
 		SQLINTEGER *, UCHAR *, SQLSMALLINT, SQLSMALLINT *, UWORD);
 
 void		logs_on_off(int cnopen, int, int);
@@ -604,7 +617,7 @@ typedef struct
 int	initialize_global_cs(void);
 enum { /* CC_conninfo_init option */
 	CLEANUP_FOR_REUSE	= 1L		/* reuse the info */
-	,COPY_GLOBALS		= (1L << 1) /* copy globals to drivers */
+	,INIT_GLOBALS		= (1L << 1) /* init globals memebers */
 };
 void	CC_conninfo_init(ConnInfo *conninfo, UInt4 option);
 void	CC_conninfo_release(ConnInfo *conninfo);
@@ -618,23 +631,6 @@ void	CC_copy_conninfo(ConnInfo *ci, const ConnInfo *sci);
 #ifdef	POSIX_THREADMUTEX_SUPPORT
 const pthread_mutexattr_t *getMutexAttr(void);
 #endif /* POSIX_THREADMUTEX_SUPPORT */
-#ifdef	UNICODE_SUPPORT
-#define WCLEN sizeof(SQLWCHAR)
-enum {
-	WCSTYPE_UNKNOWN
-	,WCSTYPE_UTF16_LE
-	,WCSTYPE_UTF32_LE
-	};
-SQLULEN	ucs2strlen(const SQLWCHAR *ucs2str);
-char	*ucs2_to_utf8(const SQLWCHAR *ucs2str, SQLLEN ilen, SQLLEN *olen, BOOL tolower);
-SQLULEN	utf8_to_ucs2_lf(const char * utf8str, SQLLEN ilen, BOOL lfconv, SQLWCHAR *ucs2str, SQLULEN buflen, BOOL errcheck);
-int	get_wcstype(void);
-int	msgtowstr(const char *, wchar_t *, int);
-int	wstrtomsg(const wchar_t *, char *, int);
-char	*wcs_to_utf8(const wchar_t *wcsstr, SQLLEN ilen, SQLLEN *olen, BOOL tolower);
-SQLULEN	utf8_to_wcs_lf(const char * utf8str, SQLLEN ilen, BOOL lfconv, wchar_t *wcsstr, SQLULEN buflen, BOOL errcheck);
-#define	utf8_to_ucs2(utf8str, ilen, ucs2str, buflen) utf8_to_ucs2_lf(utf8str, ilen, FALSE, ucs2str, buflen, FALSE)
-#endif /* UNICODE_SUPPORT */
 
 /* Define a type for defining a constant string expression */
 #ifndef	CSTR
@@ -647,6 +643,16 @@ CSTR	PRINT_NULL = "(null)";
 #define XMIN_NAME "xmin"
 #define TABLEOID_NAME "tableoid"
 
+enum
+{
+	DISALLOW_UPDATABLE_CURSORS = 0,	/* No cursors are updatable */
+	ALLOW_STATIC_CURSORS = 1L, /* Static cursors are updatable */
+	ALLOW_KEYSET_DRIVEN_CURSORS = (1L << 1), /* Keyset-driven cursors are updatable */
+	ALLOW_DYNAMIC_CURSORS = (1L << 2), /* Dynamic cursors are updatable */
+	ALLOW_BULK_OPERATIONS = (1L << 3), /* Bulk operations available */
+	SENSE_SELF_OPERATIONS = (1L << 4), /* Sense self update/delete/add */
+};
+
 #ifdef	__cplusplus
 }
 #endif
