@@ -373,25 +373,26 @@ void
 QR_add_message(QResultClass *self, const char *msg)
 {
 	char	*message = self->message;
-	size_t	alsize, pos;
+	size_t	alsize, pos, addlen;
 
 	if (!msg || !msg[0])
 		return;
+	addlen = strlen(msg);
 	if (message)
 	{
 		pos = strlen(message) + 1;
-		alsize = pos + strlen(msg) + 1;
+		alsize = pos + addlen + 1;
 	}
 	else
 	{
 		pos = 0;
-		alsize = strlen(msg) + 1;
+		alsize = addlen + 1;
 	}
 	if (message = realloc(message, alsize), NULL == message)
 		return;
 	if (pos > 0)
 		message[pos - 1] = ';';
-	strcpy(message + pos, msg);
+	strncpy_null(message + pos, msg, addlen + 1);
 	self->message = message;
 }
 
@@ -409,25 +410,26 @@ void
 QR_add_notice(QResultClass *self, const char *msg)
 {
 	char	*message = self->notice;
-	size_t	alsize, pos;
+	size_t	alsize, pos, addlen;
 
 	if (!msg || !msg[0])
 		return;
+	addlen = strlen(msg);
 	if (message)
 	{
 		pos = strlen(message) + 1;
-		alsize = pos + strlen(msg) + 1;
+		alsize = pos + addlen + 1;
 	}
 	else
 	{
 		pos = 0;
-		alsize = strlen(msg) + 1;
+		alsize = addlen + 1;
 	}
 	if (message = realloc(message, alsize), NULL == message)
 		return;
 	if (pos > 0)
 		message[pos - 1] = ';';
-	strcpy(message + pos, msg);
+	strncpy_null(message + pos, msg, addlen + 1);
 	self->notice = message;
 }
 
@@ -496,7 +498,7 @@ QR_free_memory(QResultClass *self)
 		{
 			char	plannm[32];
 
-			snprintf(plannm, sizeof(plannm), "_KEYSET_%p", self);
+			SPRINTF_FIXED(plannm, "_KEYSET_%p", self);
 			if (CC_is_in_error_trans(conn))
 			{
 				CC_mark_a_object_to_discard(conn, 's',plannm);
@@ -506,7 +508,7 @@ QR_free_memory(QResultClass *self)
 				QResultClass	*res;
 				char		cmd[64];
 
-				snprintf(cmd, sizeof(cmd), "DEALLOCATE \"%s\"", plannm);
+				SPRINTF_FIXED(cmd, "DEALLOCATE \"%s\"", plannm);
 				res = CC_send_query(conn, cmd, NULL, IGNORE_ABORT_ON_CONN | ROLLBACK_ON_ERROR, NULL);
 				QR_Destructor(res);
 			}
@@ -737,10 +739,11 @@ QR_close(QResultClass *self)
 			UDWORD		flag = 0;
 			char		buf[64];
 
+			flag = READ_ONLY_QUERY;
 			if (QR_needs_survival_check(self))
-				flag = ROLLBACK_ON_ERROR | IGNORE_ABORT_ON_CONN;
+				flag |= (ROLLBACK_ON_ERROR | IGNORE_ABORT_ON_CONN);
 
-			snprintf(buf, sizeof(buf), "close \"%s\"", QR_get_cursor(self));
+			SPRINTF_FIXED(buf, "close \"%s\"", QR_get_cursor(self));
 			/* End the transaction if there are no cursors left on this conn */
 			if (CC_is_in_trans(conn) &&
 			    CC_does_autocommit(conn) &&
@@ -749,7 +752,7 @@ QR_close(QResultClass *self)
 				mylog("QResult: END transaction on conn=%p\n", conn);
 				if ((ROLLBACK_ON_ERROR & flag) == 0)
 				{
-					strlcat(buf, ";commit", sizeof(buf));
+					STRCAT_FIXED(buf, ";commit");
 					flag |= END_WITH_COMMIT;
 					QR_set_cursor(self, NULL);
 				}
@@ -757,6 +760,7 @@ QR_close(QResultClass *self)
 					does_commit = TRUE;
 			}
 
+inolog(" !!!! %s:Case I CC_send_query %s flag=%x\n", __FUNCTION__, buf, flag);
 			res = CC_send_query(conn, buf, NULL, flag, NULL);
 			QR_Destructor(res);
 			if (does_commit)
@@ -903,9 +907,9 @@ SQLLEN	QR_move_cursor_to_last(QResultClass *self, StatementClass *stmt)
 	if (QR_once_reached_eof(self) &&
 	    self->cursTuple >= self->num_total_read)
 		return 0;
-	snprintf(movecmd, sizeof(movecmd),
+	SPRINTF_FIXED(movecmd,
 		 "move all in \"%s\"", QR_get_cursor(self));
-	res = CC_send_query(conn, movecmd, NULL, 0, stmt);
+	res = CC_send_query(conn, movecmd, NULL, READ_ONLY_QUERY, stmt);
 	if (!QR_command_maybe_successful(res))
 	{
 		QR_Destructor(res);
@@ -989,22 +993,22 @@ inolog("in total_read=%d cursT=%d currT=%d ad=%d total=%d rowsetSize=%d\n", self
 			else
 				self->cache_size = req_size;
 inolog("cache=%d rowset=%d movement=" FORMAT_ULEN "\n", self->cache_size, req_size, movement);
-			snprintf(movecmd, sizeof(movecmd),
+			SPRINTF_FIXED(movecmd,
 					 "move backward " FORMAT_ULEN " in \"%s\"",
 					 movement, QR_get_cursor(self));
 		}
 		else if (QR_is_moving_forward(self))
-			snprintf(movecmd, sizeof(movecmd),
+			SPRINTF_FIXED(movecmd,
 					 "move " FORMAT_ULEN " in \"%s\"",
 					 movement, QR_get_cursor(self));
 		else
 		{
-			snprintf(movecmd, sizeof(movecmd),
+			SPRINTF_FIXED(movecmd,
 					 "move all in \"%s\"",
 					 QR_get_cursor(self));
 			movement = INT_MAX;
 		}
-		mres = CC_send_query(conn, movecmd, NULL, 0, stmt);
+		mres = CC_send_query(conn, movecmd, NULL, READ_ONLY_QUERY, stmt);
 		if (!QR_command_maybe_successful(mres))
 		{
 			QR_Destructor(mres);
@@ -1139,7 +1143,7 @@ inolog("clear obsolete %d tuples\n", num_backend_rows);
 		RETURN(FALSE)
 
 	/* Send a FETCH command to get more rows */
-	snprintf(fetch, sizeof(fetch),
+	SPRINTF_FIXED(fetch,
 			 "fetch %d in \"%s\"",
 			 fetch_size, QR_get_cursor(self));
 
@@ -1156,7 +1160,7 @@ inolog("clear obsolete %d tuples\n", num_backend_rows);
 	qi.fetch_size = fetch_size;
 	qi.result_in = self;
 	qi.cursor = NULL;
-	res = CC_send_query(conn, fetch, &qi, 0, stmt);
+	res = CC_send_query(conn, fetch, &qi, READ_ONLY_QUERY, stmt);
 	if (!QR_command_maybe_successful(res))
 	{
 		if (!QR_get_message(self))
@@ -1398,7 +1402,7 @@ nextrow:
 						char	emsg[128];
 
 						QR_set_rstatus(self, PORES_INTERNAL_ERROR);
-						snprintf(emsg, sizeof(emsg), "Internal Error -- this_keyset == NULL ci_num_fields=%d effective_cols=%d", ci_num_fields, effective_cols);
+						SPRINTF_FIXED(emsg, "Internal Error -- this_keyset == NULL ci_num_fields=%d effective_cols=%d", ci_num_fields, effective_cols);
 						QR_set_message(self, emsg);
 						return FALSE;
 					}
