@@ -31,7 +31,6 @@
 #define INTFUNC  __stdcall
 
 extern HINSTANCE s_hModule;	/* Saved module handle. */
-extern GLOBAL_VALUES	globals;
 
 /* Constants */
 #define MIN(x,y)	  ((x) < (y) ? (x) : (y))
@@ -80,7 +79,7 @@ ConfigDSN(HWND hwnd,
 
 	/* Save original data source name */
 	if (lpsetupdlg->ci.dsn[0])
-		lstrcpy(lpsetupdlg->szDSN, lpsetupdlg->ci.dsn);
+		STRCPY_FIXED(lpsetupdlg->szDSN, lpsetupdlg->ci.dsn);
 	else
 		lpsetupdlg->szDSN[0] = '\0';
 
@@ -123,6 +122,7 @@ ConfigDSN(HWND hwnd,
 			fSuccess = FALSE;
 	}
 
+	CC_conninfo_release(&(lpsetupdlg->ci));
 	GlobalUnlock(hglbAttr);
 	GlobalFree(hglbAttr);
 
@@ -280,19 +280,16 @@ ConfigDlgProc(HWND hdlg,
 			 */
 			/* override settings in ODBC.INI */
 
-			copy_globals(&ci->drivers, &globals);
+			init_globals(&ci->drivers);
 			/* Get the rest of the common attributes */
-			getDSNinfo(ci, CONN_DONT_OVERWRITE);
-
-			/* Fill in any defaults */
-			getDSNdefaults(ci);
+			getDSNinfo(ci, lpsetupdlg->lpszDrvr);
 
 			/* Initialize dialog fields */
 			SetDlgStuff(hdlg, ci);
 	
 			/* Save drivername */
 			if (!(lpsetupdlg->ci.drivername[0]))
-				lstrcpy(lpsetupdlg->ci.drivername, lpsetupdlg->lpszDrvr);
+				STRCPY_FIXED(lpsetupdlg->ci.drivername, lpsetupdlg->lpszDrvr);
 
 			if (lpsetupdlg->fNewDSN || !ci->dsn[0])
 				ShowWindow(GetDlgItem(hdlg, IDC_MANAGEDSN), SW_HIDE);
@@ -311,6 +308,7 @@ ConfigDlgProc(HWND hdlg,
 
 			/* Process buttons */
 		case WM_COMMAND:
+			lpsetupdlg = (LPSETUPDLG) GetWindowLongPtr(hdlg, DWLP_USER);
 			switch (cmd = GET_WM_COMMAND_ID(wParam, lParam))
 			{
 					/*
@@ -334,15 +332,14 @@ ConfigDlgProc(HWND hdlg,
 					/* Accept results */
 				case IDOK:
 				case IDAPPLY:
-					lpsetupdlg = (LPSETUPDLG) GetWindowLongPtr(hdlg, DWLP_USER);
 					/* Retrieve dialog values */
 					if (!lpsetupdlg->fDefault)
 						GetDlgItemText(hdlg, IDC_DSNAME,
 									   lpsetupdlg->ci.dsn,
 									   sizeof(lpsetupdlg->ci.dsn));
+
 					/* Get Dialog Values */
 					GetDlgStuff(hdlg, &lpsetupdlg->ci);
-
 					/* Update ODBC.INI */
 					SetDSNAttributes(hdlg, lpsetupdlg, NULL);
 					if (IDAPPLY == cmd)
@@ -354,7 +351,6 @@ ConfigDlgProc(HWND hdlg,
 
 				case IDC_TEST:
 				{
-					lpsetupdlg = (LPSETUPDLG) GetWindowLongPtr(hdlg, DWLP_USER);
 					if (NULL != lpsetupdlg)
 					{
 						/* Get Dialog Values */
@@ -365,19 +361,20 @@ ConfigDlgProc(HWND hdlg,
 					break;
 				}
 				case IDC_DATASOURCE:
-					lpsetupdlg = (LPSETUPDLG) GetWindowLongPtr(hdlg, DWLP_USER);
+					/* Get Dialog Values */
+					GetDlgStuff(hdlg, &lpsetupdlg->ci);
 					DialogBoxParam(s_hModule, MAKEINTRESOURCE(DLG_OPTIONS_DRV),
 					 hdlg, ds_options1Proc, (LPARAM) &lpsetupdlg->ci);
 					return TRUE;
 
 				case IDC_DRIVER:
-					lpsetupdlg = (LPSETUPDLG) GetWindowLongPtr(hdlg, DWLP_USER);
+					/* Get Dialog Values */
+					GetDlgStuff(hdlg, &lpsetupdlg->ci);
 					DialogBoxParam(s_hModule, MAKEINTRESOURCE(DLG_OPTIONS_GLOBAL),
 						 hdlg, global_optionsProc, (LPARAM) &lpsetupdlg->ci);
 
 					return TRUE;
 				case IDC_MANAGEDSN:
-					lpsetupdlg = (LPSETUPDLG) GetWindowLongPtr(hdlg, DWLP_USER);
 					if (DialogBoxParam(s_hModule, MAKEINTRESOURCE(DLG_DRIVER_CHANGE),
 						hdlg, manage_dsnProc,
 						(LPARAM) lpsetupdlg) > 0)
@@ -434,11 +431,11 @@ ConfigDlgProc(HWND hdlg,
 #ifdef	UNICODE_SUPPORT
 #define	MESSAGEBOXFUNC		MessageBoxW
 #define	_T(str)			L ## str
-#undef	snprintf
-#define	snprintf		_snwprintf
+#define	SNTPRINTF		_snwprintf
 #else
 #define	MESSAGEBOXFUNC		MessageBoxA
 #define	_T(str)			str
+#define	SNTPRINTF		snprintf
 #endif	/* UNICODE_SUPPORT */
 
 void
@@ -468,7 +465,7 @@ test_connection(HANDLE hwnd, ConnInfo *ci, BOOL withDTC)
 	makeConnectString(out_conn, ci, sizeof(out_conn));
 mylog("conn_string=%s\n", out_conn);
 #ifdef	UNICODE_SUPPORT
-	msgtowstr(out_conn, wout_conn, _countof(wout_conn));
+	MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, out_conn, -1, wout_conn, sizeof(wout_conn) / sizeof(wout_conn[0]));
 	conn_str = wout_conn;
 #else
 	conn_str = out_conn;
@@ -481,7 +478,7 @@ mylog("conn_string=%s\n", out_conn);
 	}
 	if (!SQL_SUCCEEDED(ret = SQLSETENVATTRFUNC(env, SQL_ATTR_ODBC_VERSION, (SQLPOINTER) SQL_OV_ODBC3, 0)))
 	{
-		snprintf(szMsg, _countof(szMsg), _T("SQLAllocHandle for env error=%d"), ret);
+		SNTPRINTF(szMsg, _countof(szMsg), _T("SQLAllocHandle for env error=%d"), ret);
 		goto cleanup;
 	}
 	if (!SQL_SUCCEEDED(ret = SQLALLOCHANDLEFUNC(SQL_HANDLE_DBC, env, &conn)))
@@ -512,27 +509,31 @@ mylog("conn_string=%s\n", out_conn);
 			if (SQL_SUCCEEDED(ret))
 			{
 				SQLSETCONNECTATTRFUNC(conn, SQL_ATTR_ENLIST_IN_DTC, SQL_DTC_DONE, 0);
-				snprintf(szMsg, _countof(szMsg), _T("%s\nenlistment was successful\n"), ermsg);
+				SNTPRINTF(szMsg, _countof(szMsg), _T("%s\nenlistment was successful\n"), ermsg);
 				ermsg = szMsg;
 			}
 			else
 			{
-				SQLGETDIAGRECFUNC(SQL_HANDLE_DBC, conn, 1, NULL, &errnum, szMsg, _countof(szMsg), &str_len);
-				if (szMsg[0] != 0)
+				int	strl;
+
+				SNTPRINTF(szMsg, _countof(szMsg), _T("%s\nMSDTC error:"), ermsg);
+				for (strl = 0; strl < SQL_MAX_MESSAGE_LENGTH; strl++)
 				{
-					snprintf(szMsg, _countof(szMsg), _T("%s\nMSDTC error:%s"), ermsg, szMsg);
-					ermsg = szMsg;
+					if (!szMsg[strl])
+						break;
 				}
+				SQLGETDIAGRECFUNC(SQL_HANDLE_DBC, conn, 1, NULL, &errnum, szMsg + strl, _countof(szMsg) - strl, &str_len);
+				ermsg = szMsg;
 			}
 			CALL_ReleaseTransactionObject(pObj);
 		}
 		else if (FAILED(res))
 		{
-			snprintf(szMsg, _countof(szMsg), _T("%s\nDistibuted Transaction enlistment error %x"), ermsg, res);
+			SNTPRINTF(szMsg, _countof(szMsg), _T("%s\nDistibuted Transaction enlistment error %x"), ermsg, res);
 			ermsg = szMsg;
 		}
 #else	/* _HANDLE_ENLIST_IN_DTC_ */
-		snprintf(szMsg, _countof(szMsg), _T("%s\nDistibuted Transaction enlistment not supported by this driver"), ermsg);
+		SNTPRINTF(szMsg, _countof(szMsg), _T("%s\nDistibuted Transaction enlistment not supported by this driver"), ermsg);
 		ermsg = szMsg;
 #endif
 	}
@@ -544,8 +545,6 @@ cleanup:
 	}
 
 #undef _T
-#undef snprintf
-#define snprintf _snprintf
 
 	if (NULL != conn)
 	{
@@ -576,7 +575,7 @@ ParseAttributes(LPCSTR lpszAttributes, LPSETUPDLG lpsetupdlg)
 	size_t		cbKey;
 	char		value[MAXPGPATH];
 
-	CC_conninfo_init(&(lpsetupdlg->ci), COPY_GLOBALS);
+	CC_conninfo_init(&(lpsetupdlg->ci), INIT_GLOBALS);
 
 	for (lpsz = lpszAttributes; *lpsz; lpsz++)
 	{
@@ -611,8 +610,7 @@ ParseAttributes(LPCSTR lpszAttributes, LPSETUPDLG lpsetupdlg)
 		mylog("aszKey='%s', value='%s'\n", aszKey, value);
 
 		/* Copy the appropriate value to the conninfo  */
-		if (!copyAttributes(&lpsetupdlg->ci, aszKey, value))
-			copyCommonAttributes(&lpsetupdlg->ci, aszKey, value);
+		copyConnAttributes(&lpsetupdlg->ci, aszKey, value);
 	}
 	return;
 }
@@ -654,7 +652,7 @@ SetDSNAttributes(HWND hwndParent, LPSETUPDLG lpsetupdlg, DWORD *errcode)
 			if (SQL_SUCCESS != ret)
 			{
 				LoadString(s_hModule, IDS_BADDSN, szBuf, sizeof(szBuf));
-				wsprintf(szMsg, szBuf, lpszDSN);
+				SPRINTF_FIXED(szMsg, szBuf, lpszDSN);
 			}
 			LoadString(s_hModule, IDS_MSGTITLE, szBuf, sizeof(szBuf));
 			MessageBox(hwndParent, szMsg, szBuf, MB_ICONEXCLAMATION | MB_OK);
@@ -665,7 +663,7 @@ SetDSNAttributes(HWND hwndParent, LPSETUPDLG lpsetupdlg, DWORD *errcode)
 	}
 
 	/* Update ODBC.INI */
-	writeDriverCommoninfo(ODBC_INI, lpsetupdlg->ci.dsn, &(lpsetupdlg->ci.drivers));
+	write_Ci_Drivers(ODBC_INI, lpsetupdlg->ci.dsn, &(lpsetupdlg->ci.drivers));
 	writeDSNinfo(&lpsetupdlg->ci);
 
 	/* If the data source name has changed, remove the old name */
@@ -701,7 +699,7 @@ SetDriverAttributes(LPCSTR lpszDriver, DWORD *pErrorCode, LPSTR message, WORD cb
 		goto cleanup;
 	if (!SQLWritePrivateProfileString(lpszDriver, "ConnectFunctions", "YYN", ODBCINST_INI))
 		goto cleanup;
-	snprintf(ver_string, sizeof(ver_string), "%02x.%02x",
+	SPRINTF_FIXED(ver_string, "%02x.%02x",
 				 ODBCVER / 256,
 				 ODBCVER % 256);
 	if (!SQLWritePrivateProfileString(lpszDriver, "DriverODBCVer",
