@@ -96,6 +96,7 @@ enum {
 	,STMT_INVALID_NULL_ARG
 	,STMT_NO_RESPONSE
 	,STMT_COMMUNICATION_ERROR
+	,STMT_STRING_CONVERSION_ERROR
 };
 
 /* statement types */
@@ -259,15 +260,14 @@ struct StatementClass_
 	po_ind_t	prepare;	/* is this a prepared statement ? */
 	po_ind_t	prepared;	/* is this statement already
 					 * prepared at the server ? */
-	po_ind_t	internal;	/* Is this statement being called
-							 * internally ? */
+	po_ind_t	external;	/* Allocated via SQLAllocHandle() */
 	po_ind_t	transition_status;	/* Transition status */
 	po_ind_t	multi_statement; /* -1:unknown 0:single 1:multi */
-	po_ind_t	rbonerr;	/* rollback on error */
+	po_ind_t	rb_or_tc;	/* rollback on error */
 	po_ind_t	discard_output_params;	 /* discard output parameters on parse stage */
 	po_ind_t	cancel_info;	/* cancel information */
 	po_ind_t	ref_CC_error;	/* refer to CC_error ? */
-	po_ind_t	lock_CC_for_rb;	/* lock CC for statement rollback ? */
+//	po_ind_t	lock_CC_for_rb;	/* lock CC for statement rollback ? */
 	po_ind_t	join_info;	/* have joins ? */
 	po_ind_t	parse_method;	/* parse_statement is forced or ? */
 	po_ind_t	curr_param_result; /* current param result is set ? */
@@ -281,6 +281,7 @@ struct StatementClass_
 	SQLLEN		exec_current_row;
 
 	unsigned char	miscinfo;
+	unsigned char	execinfo;
 	po_ind_t	updatable;
 	SQLLEN		diag_row_count;
 	char		*load_statement; /* to (re)load updatable individual rows */
@@ -289,6 +290,7 @@ struct StatementClass_
 	ssize_t		where_pos;
 	SQLLEN		last_fetch_count_include_ommitted;
 	time_t		stmt_time;
+	struct tm	localtime;
 	/* SQL_NEED_DATA Callback list */
 	StatementClass	*execute_delegate;
 	StatementClass	*execute_parent;
@@ -423,10 +425,14 @@ enum
 #define SC_set_fetchcursor(a)	((a)->miscinfo |= (1L << 1))
 #define SC_no_fetchcursor(a)	((a)->miscinfo &= ~(1L << 1))
 #define SC_is_fetchcursor(a)	(((a)->miscinfo & (1L << 1)) != 0)
-#define SC_set_with_hold(a)	((a)->miscinfo |= (1L << 3))
-#define SC_set_without_hold(a)	((a)->miscinfo &= ~(1L << 3))
-#define SC_is_with_hold(a)	(((a)->miscinfo & (1L << 3)) != 0)
-#define SC_miscinfo_clear(a)	((a)->miscinfo &= (1L << 3))
+#define SC_miscinfo_clear(a)	((a)->miscinfo = 0)
+#define SC_set_with_hold(a)	((a)->execinfo |= 1L)
+#define SC_set_without_hold(a)	((a)->execinfo &= (~1L))
+#define SC_is_with_hold(a)	(((a)->execinfo & 1L) != 0)
+#define SC_set_readonly(a)	((a)->execinfo |= (1L << 1))
+#define SC_set_no_readonly(a)	((a)->execinfo &= ~(1L << 1))
+#define SC_is_readonly(a)	(((a)->execinfo & (1L << 1)) != 0)
+#define SC_execinfo_clear(a)	(((a)->execinfo = 0)
 #define	STMT_HAS_OUTER_JOIN	1L
 #define	STMT_HAS_INNER_JOIN	(1L << 1)
 #define SC_has_join(a)		(0 != (a)->join_info)
@@ -435,15 +441,10 @@ enum
 #define SC_set_outer_join(a)	((a)->join_info |= STMT_HAS_OUTER_JOIN)
 #define SC_set_inner_join(a)	((a)->join_info |= STMT_HAS_INNER_JOIN)
 
-#define SC_start_stmt(a)	((a)->rbonerr = 0)
-#define SC_start_tc_stmt(a)	((a)->rbonerr = (1L << 1))
-#define SC_is_tc_stmt(a)	(((a)->rbonerr & (1L << 1)) != 0)
-#define SC_start_rb_stmt(a)	((a)->rbonerr = (1L << 2))
-#define SC_is_rb_stmt(a)	(((a)->rbonerr & (1L << 2)) != 0)
-#define SC_set_accessed_db(a)	((a)->rbonerr |= (1L << 3))
-#define SC_accessed_db(a)	(((a)->rbonerr & (1L << 3)) != 0)
-#define SC_start_rbpoint(a)	((a)->rbonerr |= (1L << 4))
-#define SC_started_rbpoint(a)	(((a)->rbonerr & (1L << 4)) != 0)
+#define SC_start_tc_stmt(a)	((a)->rb_or_tc = (1L << 1))
+#define SC_is_tc_stmt(a)	(((a)->rb_or_tc & (1L << 1)) != 0)
+#define SC_start_rb_stmt(a)	((a)->rb_or_tc = (1L << 2))
+#define SC_is_rb_stmt(a)	(((a)->rb_or_tc & (1L << 2)) != 0)
 #define SC_unref_CC_error(a)	(((a)->ref_CC_error) = FALSE)
 #define SC_ref_CC_error(a)	(((a)->ref_CC_error) = TRUE)
 #define SC_can_parse_statement(a) (STMT_TYPE_SELECT == (a)->statement_type)
@@ -452,7 +453,7 @@ enum
  * it's not currently supported with array-bound parameters.
  */
 #define SC_may_use_cursor(a) \
-	(SC_get_APDF(a)->paramset_size <= 1 &&								\
+	(SC_get_APDF(a)->paramset_size <= 1 &&	\
 	 (STMT_TYPE_SELECT == (a)->statement_type || STMT_TYPE_WITH == (a)->statement_type) )
 #define SC_may_fetch_rows(a) (STMT_TYPE_SELECT == (a)->statement_type || STMT_TYPE_WITH == (a)->statement_type)
 
@@ -509,6 +510,7 @@ RETCODE		SC_fetch(StatementClass *self);
 void		SC_free_params(StatementClass *self, char option);
 void		SC_log_error(const char *func, const char *desc, const StatementClass *self);
 time_t		SC_get_time(StatementClass *self);
+struct tm	*SC_get_localtime(StatementClass *self);
 SQLLEN		SC_get_int4_bookmark(StatementClass *self);
 RETCODE		SC_pos_reload(StatementClass *self, SQLULEN index, UInt2 *, Int4);
 RETCODE		SC_pos_update(StatementClass *self, SQLSETPOSIROW irow, SQLULEN index, const KeySet *keyset);
@@ -536,7 +538,7 @@ int		enqueueNeedDataCallback(StatementClass *self, NeedDataCallfunc, void *);
 RETCODE		dequeueNeedDataCallback(RETCODE, StatementClass *self);
 void		cancelNeedDataState(StatementClass *self);
 int		StartRollbackState(StatementClass *self);
-RETCODE		SetStatementSvp(StatementClass *self);
+RETCODE		SetStatementSvp(StatementClass *self, unsigned int option);
 RETCODE		DiscardStatementSvp(StatementClass *self, RETCODE, BOOL errorOnly);
 
 QResultClass *ParseAndDescribeWithLibpq(StatementClass *stmt, const char *plan_name, const char *query_p, Int2 num_params, const char *comment, QResultClass *res);
